import inspect
import random
import logging
import time
import datetime
import calendar
import sched

import mini_buildd.misc

LOG = logging.getLogger(__name__)


class _Job():
    def __enter(self):
        self._next_run = self._next_run_func()
        self.event = self.scheduler.enter((self._next_run - datetime.datetime.now()).total_seconds(), 1,
                                          self.__action, argument=self._argument, kwargs=self._kwargs)

    def __init__(self, scheduler, next_run_func, action, argument=None, kwargs=None):
        self.scheduler = scheduler
        self._action = action
        self._argument = argument or ()
        self._kwargs = kwargs or {}

        self._next_run_func = next_run_func
        self.__enter()

    def __action(self, *args, **kwargs):
        LOG.info("Running cronjob: {inspect.getdoc(self._action)}")
        self._action(*args, **kwargs)
        self.__enter()

    def __str__(self):
        return f"Cronjon({inspect.getdoc(self._action)}): Next run on {self._next_run}."


class Tab(mini_buildd.misc.Thread):
    @classmethod
    def __rdate(cls, base_date, h_start, h_stop):
        return datetime.datetime(year=base_date.year, month=base_date.month, day=base_date.day,
                                 hour=random.randrange(h_start, h_stop),
                                 minute=random.randrange(60), second=random.randrange(60))

    @classmethod
    def next_monthly_run(cls):
        """First day of next month between 0am and 1am."""
        today = datetime.date.today()
        return cls.__rdate(today.replace(day=1) + datetime.timedelta(calendar.monthrange(today.year, today.month)[1]), 0, 1)

    @classmethod
    def next_weekly_run(cls):
        """Next sunday between 2am and 3am."""
        now = datetime.datetime.now()
        return cls.__rdate(now + datetime.timedelta(days=7 - (now.isoweekday() % 7)), 2, 3)

    @classmethod
    def next_daily_run(cls):
        """Next day between 4am and 6am."""
        return cls.__rdate(datetime.datetime.now() + datetime.timedelta(days=1), 4, 6)

    def __init__(self):
        super().__init__(name="crontab")
        self.scheduler = sched.scheduler()
        self.jobs = []

    def add(self, *args, **kwargs):
        self.jobs.append(_Job(self.scheduler, *args, **kwargs))

    def add_monthly(self, *args, **kwargs):
        self.add(self.next_monthly_run, *args, **kwargs)

    def add_weekly(self, *args, **kwargs):
        self.add(self.next_weekly_run, *args, **kwargs)

    def add_daily(self, *args, **kwargs):
        self.add(self.next_daily_run, *args, **kwargs)

    def mbd_run(self):  # pylint: disable=inconsistent-return-statements
        nxt = self.scheduler.run(blocking=False)
        if nxt and not self.in_shutdown():
            # self.log_status()
            time.sleep(min(nxt, 5))
            return self.RUN_AGAIN

    def log_status(self):
        for job in self.jobs:
            LOG.debug(f"{job}")
